ROS Region Detection
--------------------

The region_finder package provides a mechanism to dynamically build 2D
regions from a SLAM map. It generates regions by trying to detect
doorway-like features and segregating a space based on these doorways.

Regions that are separated by a doorway are guaranteed to overlap and
therefore the constructed regions can be used to provide high-level
path planning based on regions.

The only input to the region_finder is an occupancy grid (e.g., as generated by
the gmapping package). The output of the region_finder is a set of
regions. There are two versions of the region_finder that vary in how they
output the region data:

- region_finder: This version publishes the full region list (with all
  region data) at every pass.

- region_finder_serv (experimental): This version publishes only the
region id list at every pass. The user can then query a service for
only the regions that have changed.

By default, it reads from the "/map" topic and writes to the
"/regions" topic.

As a debugging aid, there is a basic GUI mode to make the region_finder display
the detected regions as well as the intermediate maps used to compute the
regions:

   rosrun region_finder region_finder _gui:=true

Packages
--------

- region_finder: Main package with the region_finder node.
- region_msgs: The various ROS messages.
- region_viewer: A python based script to provide basic viewing of regions.

Dependencies
------------

- region_finder:
    - Usual ROS C++ libraries.
    - CGAL: C++ computational geometry library.
    - OpenCV.

- region_viewer:
    - Usual ROS python libraries.
    - OpenCV - cv library.


Example usage:
--------------

1: Building from static map
---------------------------

There is an example static map of a (Player-Stage) simulated home. This map is
turned into an occupancy grid using the standard ROS map_server package. To run
the map:

    roslaunch region_finder static_map.launch

Now run the region_finder to generate the regions. I've turned on the GUI mode
to display the regions and intermediate maps:

   roslaunch region_finder region_finder.launch

You should see a number of graphics:

- The original occupancy grid.

- A filtered occupancy grid that treats unknown space (that is connected to
  occupied space) as occupied space.

- A map containing the voronoi points and identified "doorways". The voronoi
  points are generated from the filtered occupancy grid. Then the voronoi points
  are analysed to identify the doorway-like regions.

- A map containing the overlapping regions. This is generated from the voronoi
  points and doorway points.

The regions (including the voronoi points associated with each region) are
published on /regions. There is a simple python program that shows how you can
use these regions.

    roslaunch region_viewer region_viewer.launch


2: Building dynamically
-----------------------

There is an example rosbag that contains an occupancy grid (for the stage
simulated home) being built dynamically over time. To see the regions being
detected and modified dynamically run the following:

   roslaunch region_finder region_finder.launch

And in another terminal:

  rosbag play regin_finder/bags/stage_at_home_regions.bag
