/*
 * basic_map_updater.h
 *
 * Update the basic map based on the occupancy grid. Tries to be smart about
 * converting unknown space to either free or occupied space depending on what
 * is around it.
 *
 *  Created on: 24/10/2011
 *      Author: daver
 */

#ifndef BASIC_MAP_UPDATER_H_
#define BASIC_MAP_UPDATER_H_

#include "grid2d.h"
#include "basic_map.h"
#include "mapmetadata_tools.h"
#include "bushfire2.h"

class BasicMapUpdater
{
  struct OccGrid2BasicMap
  {
    int threshold_;

    inline int operator()(int val)
    {
      if (val < 0){ return -1; }
      else if (val >= threshold_){ return 100; }
      return 0;
    }
    OccGrid2BasicMap(int threshold): threshold_(threshold){}
  } og2bm_fn_;

  typedef BushFire2<int, IntPoint2D> t_bushfire;
  t_bushfire propagator_;
  t_basicmap *basicmap_;
  std::auto_ptr<t_occgrid> occgrid_;
  std::vector<IntPoint2D> adjacents_;

  t_pointset changes_;
  t_pointset tochange_;
  t_pointset visited_;
  void run(t_basicmap& basicmap, const t_occgrid& occgrid);
  void runFillOccupied(t_pointset& doneset);
  void runFillFree(t_pointset& doneset);

public:
  // Callback functions. Not to be called by an external program
  void fillOccupied(t_bushfire& bf, const int& val, const IntPoint2D& p);
  void fillFree(t_bushfire& bf, const int& val, const IntPoint2D& p);

  BasicMapUpdater(int threshold = 50);

  template<typename It>
  void update(t_basicmap& basicmap, const t_occgrid& occgrid, It it, It end)
  {
    check_matching_shape(occgrid, basicmap);
    tochange_.clear();
    changes_.clear();
    visited_.clear();
    while (it != end)
    {

     /*
      int pv = og2bm_fn_(get_cell(occgrid, *it));
      BasicMapCell& pbmc = get_cell(basicmap, *it);
      // If the cell hasn't changed and then do nothing
      if ((pbmc.occupancy_ == BasicMapCell::occupied && pv == 1) ||
          (pbmc.occupancy_ == BasicMapCell::free && pv == 0))
      {
        continue;
      }
      else
      {*/
      tochange_.insert(*it);
      ++it;
    }
    run(basicmap, occgrid);
  }

  void update(t_basicmap& basicmap, const t_occgrid& occgrid);
  const t_pointset& changes() const { return changes_; }
};


#endif /* BASIC_MAP_UPDATER_H_ */
