/*
 * basic_map_updater.cpp
 *
 *  Created on: 24/10/2011
 *      Author: daver
 */

#include "basic_map_updater.h"

BasicMapUpdater::BasicMapUpdater(int threshold): og2bm_fn_(threshold)
{

}

void BasicMapUpdater::update(t_basicmap& basicmap, const t_occgrid& occgrid)
{
  std::vector<IntPoint2D> points;
  all_points(occgrid, points);
  update(basicmap, occgrid, points.begin(), points.end());
}

void BasicMapUpdater::runFillOccupied(t_pointset& doneset)
{
  propagator_.clear();
  // Fill unknown space connected to occupied space as occupied
  BOOST_FOREACH(const IntPoint2D& p, tochange_)
  {
    if (doneset.find(p) != doneset.end()) continue;
    int pv = og2bm_fn_(get_cell(*occgrid_, p));

    if (pv == 0) continue; // Only consider occupied and unknown cells
    if (pv == 100) // Propagate out from occupied cells
    {
      propagator_.push(0, p);
      doneset.insert(p);
      continue;
    }

    // An unknown cell so look for occupied adjacents
    adjacents_.clear();
    adjacents8(*occgrid_, p, adjacents_);
    BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
    {
      if (doneset.find(ap) != doneset.end()) continue;
      if (og2bm_fn_(get_cell(*occgrid_, ap)) == 100)
      {
        propagator_.push(0, ap);
        doneset.insert(p);
        doneset.insert(ap);
      }
    }
  }
  std::cout << "Starting fillOccupied!!! " << std::endl;
  propagator_.update(t_bushfire::make_callback(this, &BasicMapUpdater::fillOccupied));
  std::cout << "Finished fillOccupied!!! " << std::endl;
}

void BasicMapUpdater::runFillFree(t_pointset& doneset)
{
  propagator_.clear();
  // Fill any remaining unknown space as free
  BOOST_FOREACH(const IntPoint2D& p, tochange_)
  {
    if (doneset.find(p) != doneset.end()) continue;
    doneset.insert(p);
    int pv = og2bm_fn_(get_cell(*occgrid_, p));
    if (pv == 100) continue;
    else if (pv == 0)
    {
      propagator_.push(0, p);
      continue;
    }
    // An unknown cell so look for free adjacents
    adjacents_.clear();
    adjacents8(*occgrid_, p, adjacents_);
    BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
    {
      if (doneset.find(ap) != doneset.end()) continue;
      if (og2bm_fn_(get_cell(*occgrid_, ap)) == 0)
      {
        propagator_.push(0, ap);
        doneset.insert(ap);
      }
    }
  }

  std::cout << "Starting fillFree!!! " << std::endl;
  propagator_.update(t_bushfire::make_callback(this, &BasicMapUpdater::fillFree));
  std::cout << "Finished fillFree!!! " << std::endl;

}


void BasicMapUpdater::run(t_basicmap& basicmap, const t_occgrid& occgrid)
{
  basicmap_ = &basicmap;
  occgrid_ = create_grid2d<t_occgrid, t_occgrid>(occgrid);
  t_pointset doneset;

  runFillOccupied(doneset); // Fill unknown spaces connected to occupied cells
  runFillFree(doneset);     // Fill remaining unknown spaces as free
}

void BasicMapUpdater::fillOccupied(t_bushfire& bf, const int& val, const IntPoint2D& p)
{
  if (visited_.find(p) != visited_.end()) return; // already visited the cell
  visited_.insert(p);
  BasicMapCell& pbmc = get_cell(*basicmap_, p);
  int pogc = og2bm_fn_(get_cell(*occgrid_, p));
  if (pogc == 100 && pbmc.isOccupied()) return; // Cell is already occupied so do nothing
  BOOST_ASSERT(pogc != 0); // This shouldn't be called for a free cell
  if (pbmc.isFree()) changes_.insert(p);
  pbmc.occupancy_ = BasicMapCell::occupied;
  get_cell(*occgrid_, p) = 100;

  // Propagate to adjacent cells
  adjacents_.clear();
  adjacents8(*occgrid_, p, adjacents_);
  BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
  {
    if (visited_.find(ap) != visited_.end()) continue;
    if (get_cell(*occgrid_, ap) == -1) propagator_.push(val+1, ap);
  }
}

void BasicMapUpdater::fillFree(t_bushfire& bf, const int& val, const IntPoint2D& p)
{
  if (visited_.find(p) != visited_.end()) return; // already visited the cell
  visited_.insert(p);
  BasicMapCell& pbmc = get_cell(*basicmap_, p);
  int pogc = og2bm_fn_(get_cell(*occgrid_, p));
  if (pogc == 0 && pbmc.isFree()) return; // Cell is already free so do nothing
  BOOST_ASSERT(pogc != 100); // This function shouldn't be called for an occupied cell
  if (pbmc.isOccupied()) changes_.insert(p);
  pbmc.occupancy_ = BasicMapCell::free;
  get_cell(*occgrid_, p) = 0;

  // Propagate to adjacent cells
  adjacents_.clear();
  adjacents8(*occgrid_, p, adjacents_);
  BOOST_FOREACH(const IntPoint2D& ap, adjacents_)
  {
    if (visited_.find(ap) != visited_.end()) continue;
    if (get_cell(*occgrid_, ap) == -1) propagator_.push(val+1, ap);
  }
}


