/*
 * connected_points.cpp
 *
 *  Created on: 30/09/2011
 *      Author: daver
 */


/*****************************************************************************/

#include <utility>
#include <boost/tuple/tuple.hpp>
#include <boost/property_map/property_map.hpp>
#include "connected_points.h"

ConnectedPoints::t_vertex ConnectedPoints::getVertex(const IntPoint2D& p)
{
  t_p2v::iterator it = p2v_.find(p);
  BOOST_ASSERT(it != p2v_.end());
  return it->second;
}


void ConnectedPoints::condRemoveVertex(const IntPoint2D& p)
{
  t_graph::vertex_iterator vi, viend;
  boost::tie(vi, viend) = boost::vertices(graph_);

  while (vi != viend)
  {
    t_graph::vertex_descriptor v = *vi++;
    if (graph_[v].point == p)
    {
      boost::clear_vertex(v, graph_);
      boost::remove_vertex(v, graph_);
      return;
    }
  }
  /*
  t_p2v::iterator it = p2v_.find(p);
  if (it != p2v_.end())
  {
    t_vertex v = it->second;
    boost::clear_vertex(v, graph_);
    boost::remove_vertex(v, graph_);
    p2v_.erase(it);
//    v2p_.erase(v) ;
  }
  */
}

ConnectedPoints::t_vertex ConnectedPoints::condAddVertex(const IntPoint2D& p)
{
  t_p2v::iterator it = p2v_.find(p);
  if (it == p2v_.end())
  {
    t_vertex v = boost::add_vertex(Vertex(p), graph_);
    p2v_[p] = v;
//    v2p_[v] = p;
    return v;
  }
  return it->second;
}


void ConnectedPoints::buildPoint2VertexMap()
{
  p2v_.clear();
  //t_graph::vertex_iterator vi, viend;
  typedef boost::graph_traits<t_graph>::vertex_iterator t_vi;
  t_vi vi, viend;
  boost::tie(vi, viend) = boost::vertices(graph_);
  while (vi != viend)
    p2v_[graph_[*vi].point] = *(vi++);
}


// A euclidean distance heuristic
class distance_heuristic : public boost::astar_heuristic<ConnectedPoints::t_graph, float>
{
public:
  typedef ConnectedPoints::t_graph t_graph;
  typedef t_graph::vertex_descriptor t_vertex;

  distance_heuristic(t_graph graph, t_vertex goal) :
    graph_(graph), goal_(graph[goal].point)
  {
    //goal_ = graph[goal].point;
  }
  float operator()(t_vertex v)
  {
    return goal_.distanceTo(graph_[v].point);
  }
private:
  t_graph graph_;
  const IntPoint2D& goal_;
};



struct found_goal {}; // exception for termination
struct bounds_exceeded {}; // exception for when there is no path within the bounds

// visitor that terminates when we find the goal
template<class Vertex>
class astar_goal_visitor : public boost::default_astar_visitor
{
  Vertex goal_;
  IntPoint2D goalP_;
  float threshold_;
public:
  astar_goal_visitor(Vertex goal, const IntPoint2D& goalP, float threshold) :
    goal_(goal), goalP_(goalP), threshold_(threshold) {}
  template <class Graph>
  void examine_vertex(Vertex u, Graph& g) {
    if(u == goal_)
      throw found_goal();
//    else if (g[u].point.distanceTo(goalP_) > threshold_)
//      throw bounds_exceeded();
  }
};

void ConnectedPoints::buildVertexIndexMap(t_v2i& v2i)
{
  t_graph::vertex_iterator vi, viend;
  boost::tie(vi, viend) = boost::vertices(graph_);
  size_t i=-1;
  while (vi != viend)
    v2i[*vi++] = ++i;
}

bool ConnectedPoints::isBoundedlyConnected(const IntPoint2D& a, const IntPoint2D& b)
{
  t_v2f v2f;
  t_v2v v2v;
  t_v2i v2i;

  boost::associative_property_map<t_v2f> dmap(v2f);
  boost::associative_property_map<t_v2v> pmap(v2v);
  boost::associative_property_map<t_v2i> vimap(v2i);
  t_vertex va = getVertex(a);
  t_vertex vb = getVertex(b);
  buildVertexIndexMap(v2i);



  try {
    boost::astar_search(graph_, va,
                        distance_heuristic(graph_, vb),
                        boost::weight_map(boost::get(&Edge::weight, graph_)).
                        vertex_index_map(vimap).
                        predecessor_map(pmap).
                        distance_map(dmap).
                        visitor(astar_goal_visitor<t_vertex>(vb, b, threshold_)));


   }
   catch(found_goal& fg)
   { // found a path to the goal
/*
     std::list<t_vertex> shortest_path;
     for(t_vertex v = vb;; v = pmap[v]) {
       shortest_path.push_front(v);
       if(pmap[v] == v)
         break;

     }
     */
     if (dmap[vb] > threshold_)
     {
       std::cout << "No shortest (TE) path less than " << threshold_ << " from " << a << " to " << b << ": "
                 << dmap[vb] << std::endl;
       return false;
     }

     std::cout << "Shortest path less than " << threshold_ << " from " << a << " to " << b << ": ";
/*
     std::list<t_vertex>::iterator spi = shortest_path.begin();
     for(++spi; spi != shortest_path.end(); ++spi)
       std::cout << " -> " << graph_[*spi].point;
       */
     std::cout << dmap[vb] << std::endl;
     return true;
   }
   catch(bounds_exceeded& )
   {
     std::cout << "No shortest path (BE) less than " << threshold_ << " from " << a << " to " << b << std::endl;
     return false;
   }
   // Should only get here if are not using bounds_exceeded and there is really no path between the two points
   std::cout << "There really is no path from " << a << " to " << b << std::endl;
   BOOST_ASSERT(true);
   return false;
}


